/**
 *  FWHID_HandMotion.c
 *  FingerWorks HID Hand Motion SDK (Mac OS X Implementation)
 *
 *  Created by Wayne Westerman on Wed Jun 25 2003.
 *  Last Updated by Wayne Westerman on Wednesday Feb 11 2004.
 *  Copyright (c) 2003-2004 FingerWorks, Inc. All rights reserved.
 **/
 /* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is FWHID_HandMotion.c,
 *   Mac OS X Implementation of FingerWorks HID Hand Motion SDK.
 *
 * The Initial Developer of the Original Code is
 * Wayne Westerman, FingerWorks, Inc.
 *
 * Portions created by the Initial Developer are Copyright (C) 2003-2004
 * by FingerWorks Inc. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */


// system includes ----------------------------------------------------------


#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

// project includes ---------------------------------------------------------
#include "FWHID_HandMotion.h"
#include "fwhid_HandMotionInternals.h"

//pre-compilation options ---------------------------------------------------

// prototypes ---------------------------------------------------------------

//globals (to this file)
static FWHIDHandEventCALLBACK gHandeventCallbackp = NULL;

// functions (external but private) ----------------------------------------------


void fwhand_clearMotion(FWHID_HandMotionPtr pHandMotion) {
    if(pHandMotion != NULL) {
        pHandMotion->lastPressure = pHandMotion->handPressure;
        pHandMotion->lastFingers.word = pHandMotion->currentFingers.word;
		pHandMotion->handActionCode = 0;
		pHandMotion->slideAxesCode = 0;
		//pHandMotion->handPositionX = 0; //shouldn't clear because Mac OS X won't send element update if position didn't change
		//pHandMotion->handPositionY = 0; //shouldn't clear because Mac OS X won't send element update if position didn't change
        pHandMotion->Xtrans = 0;
        pHandMotion->Ytrans = 0;
        pHandMotion->Ztrans = 0;
        pHandMotion->Xrot = 0;
        pHandMotion->Yrot = 0;
        pHandMotion->Zrot = 0;
        pHandMotion->Xscale = 0;
        pHandMotion->Yscale = 0;
        pHandMotion->Zscale = 0;
    }
}

int fwhand_getMotionType(FWHID_HandMotionPtr pHandMotion) {
    if(pHandMotion != NULL) {
        bool has_translation = (pHandMotion->Xtrans != 0 || pHandMotion->Ytrans != 0 || pHandMotion->Ztrans != 0);
        bool has_rotation = (pHandMotion->Xrot != 0 || pHandMotion->Yrot != 0 || pHandMotion->Zrot != 0);
        bool has_scale = (pHandMotion->Xscale != 0 || pHandMotion->Yscale != 0 || pHandMotion->Zscale != 0);
        bool pressure_change = (pHandMotion->handPressure != pHandMotion->lastPressure);
		bool has_action_code = (pHandMotion->handActionCode != 0);
		if(has_action_code) {
			return kEventHandCommand;
		} else if(!has_translation && !has_scale && !has_rotation) {
            if(pressure_change)
                return kEventHandPressure;
            else return kEventHandNoChange;
        } else {
            return kEventHandPressure | (has_translation | (has_rotation << 1) | (has_scale << 2));
        }
    }
    return kEventHandNoChange;
}

void fwhand_processAccumulatedMotion(FWHID_HandMotionPtr pHandMotion) {
    if(pHandMotion != NULL) {
        int handEventID = fwhand_getMotionType(pHandMotion);
        if(handEventID != kEventHandNoChange && gHandeventCallbackp != NULL) {
            (*gHandeventCallbackp)(handEventID,pHandMotion);
        }
        fwhand_clearMotion(pHandMotion);
    }
}

void fwhand_processChordChange(FWHID_HandMotionPtr pHandMotion, unsigned new_chord_word) {
    FWHID_ChordSpecifier new_chord;
    new_chord.word = new_chord_word;
    if(pHandMotion != NULL) {
        pHandMotion->lastChord.word = pHandMotion->selectedChord.word;
        if(new_chord.compare.chord == 0) {
            fwhand_processAccumulatedMotion(pHandMotion); //forward any accumulated motion before clearing chord selection and sending release
            pHandMotion->selectedChord.word = new_chord.word; //this should clear chord selection
            if(gHandeventCallbackp != NULL) {
                (*gHandeventCallbackp)(kEventHandChordReleased,pHandMotion);  //send release event
            }
        } else {
            FWHID_HandMotion tempzeroed_motion;
            pHandMotion->selectedChord.word = new_chord.word;
            tempzeroed_motion = *pHandMotion;
            fwhand_clearMotion(&tempzeroed_motion); //make sure no beginning motion is included in chord selection event
            if(gHandeventCallbackp != NULL) {
                (*gHandeventCallbackp)(kEventHandChordSelected,&tempzeroed_motion);
            }
            fwhand_processAccumulatedMotion(pHandMotion); //now forward any beginning motion
        }
    }
}

static FWHID_HandMotion gRightHandMotion;
static FWHID_HandMotion gLeftHandMotion;


static unsigned fwhand_bytePairToUnsigned(FWBYTE *bytearray, int offset) {
	return (bytearray[offset] & 0xFF) | ((bytearray[offset+1] & 0xFF) << 8);
}
/*  NOTE:  Mac OS X implementation does not use this raw parsing
    function since data comes in already broken up into HID collections elements */
void fwhand_parseRawHandMotion(FWBYTE *bytearray, int numbytes, int firmware_version) {
	FWHID_HandMotionPtr pHandMotion;
	if(bytearray != NULL && numbytes >= 15) {
		unsigned newchord;
		int evenize_alignment = (firmware_version > 0x141);

		if(bytearray[0] == kFWHID_RightHandMotion_ReportID) {
			pHandMotion = &gRightHandMotion;
		} else if(bytearray[0] == kFWHID_LeftHandMotion_ReportID) {
			pHandMotion = &gLeftHandMotion;
		} else return;

		if(evenize_alignment) {
			pHandMotion->handActionCode = bytearray[1] & 0xFF;
			pHandMotion->slideAxesCode = bytearray[5] & 0xFF;
		} else {
			pHandMotion->handActionCode = bytearray[4] & 0xFF;
			if(numbytes >= 16)
				pHandMotion->slideAxesCode = bytearray[15] & 0xFF;
		}

		newchord = fwhand_bytePairToUnsigned(bytearray,1+evenize_alignment);
		if(newchord != pHandMotion->selectedChord.word)
			fwhand_processChordChange(pHandMotion,newchord);
		else pHandMotion->lastChord.word = pHandMotion->selectedChord.word;
		
        pHandMotion->lastFingers.word = pHandMotion->currentFingers.word;
		pHandMotion->currentFingers.word = bytearray[3+evenize_alignment] & 0xFF;
        pHandMotion->lastPressure = pHandMotion->handPressure;
		pHandMotion->handPressure = fwhand_bytePairToUnsigned(bytearray,5+evenize_alignment);
		pHandMotion->handPositionX = (float)(((signed short)fwhand_bytePairToUnsigned(bytearray,7+evenize_alignment))/400.0); //sign-extend and convert from 25 um to cm
		pHandMotion->handPositionY = (float)(((signed short)fwhand_bytePairToUnsigned(bytearray,9+evenize_alignment))/400.0); //sign-extend and convert from 25 um to cm
        pHandMotion->Xtrans = (signed char)bytearray[11+evenize_alignment];//sign-extend
        pHandMotion->Ytrans = (signed char)bytearray[12+evenize_alignment];//sign-extend
        pHandMotion->Ztrans = (signed char)bytearray[13+evenize_alignment];//sign-extend
        pHandMotion->Xrot = 0;
        pHandMotion->Yrot = 0;
        pHandMotion->Zrot = (signed char)bytearray[14+evenize_alignment];//sign-extend
        pHandMotion->Xscale = 0;
        pHandMotion->Yscale = 0;
        pHandMotion->Zscale = 0;
		
		fwhand_processAccumulatedMotion(pHandMotion);
	}
}

void fwhand_clearState(void) {
    memset(&gRightHandMotion,0,sizeof(FWHID_HandMotion));
    memset(&gLeftHandMotion,0,sizeof(FWHID_HandMotion));
}


// functions (external/public) ----------------------------------------------

void FWHID_setHandEventCallback(FWHIDHandEventCALLBACK fwcallback) {
    gHandeventCallbackp = fwcallback;
}

FWHIDHandEventCALLBACK FWHID_getHandEventCallback(void) {
    return gHandeventCallbackp;
}

//the default FWHIDHandEventCALLBACK
void FWHID_printHandMotion(int handEventID, FWHID_HandMotionPtr pHandMotion) {
	char printbuf[256];
	FWHID_snprintfHandMotion(printbuf,256,handEventID,pHandMotion);
	printf("%s",printbuf);
    fflush(stdout);	
}

int FWHID_snprintfHandMotion(char *buffer, size_t bufsize, int handEventID, FWHID_HandMotionPtr pHandMotion) {
	char actioneventbuf[40];
	char *bimanualname = "";
	char *handname = "";
	char *handeventname = "";
    char *handevent_typenames[] = {
        "NullHandEvent ",
        "ChordSelect   ",
        "ChordRelease  ",
        "...reserved...",
        "ActionCmd",
        "...reserved...",
        "...reserved...",
        "...reserved...",
        "Pressure      ",
        "Translate     ",
        "Rotate        ",
        "Transl+Rotate ",
        "Scale         ",
        "Transl+Scale  ",
        "Rotate+Scale  ",
        "Motion        "
    };

    if(pHandMotion->selectedChord.flags.bimanualSync) {
        bimanualname = "Bimanual ";
    }

    if(pHandMotion->selectedChord.flags.lefthand) {
        handname = "Left H";
    } else if(pHandMotion->selectedChord.flags.righthand) {
        handname = "RightH";
    } else handname = "Null H";

    if(handEventID == kEventHandCommand) {
		snprintf(actioneventbuf,40,"%s 0x%x",
			handevent_typenames[handEventID],
			pHandMotion->handActionCode);
		handeventname = actioneventbuf;
	} else if(handEventID >= 0 && handEventID < kEventHandIDmax) {
        handeventname = handevent_typenames[handEventID];
    } else handeventname = "UnknownEvent ";

    /*    printf(": Schord=0x%x Czone=0x%x pres=%u X=%d Y=%d Z=%d Rz=%d Ry=%d Rz=%d Sx=%d Sy=%d Sz=%d\n",
        pHandMotion->selectedChord.word, pHandMotion->currentZone.word,pHandMotion->handPressure,
        pHandMotion->Xtrans,pHandMotion->Ytrans,pHandMotion->Ztrans,
        pHandMotion->Xrot,pHandMotion->Xrot,pHandMotion->Zrot,
        pHandMotion->Xscale,pHandMotion->Yscale,pHandMotion->Zscale);*/
    return snprintf(buffer,bufsize,"%s%s %s: Schrd=0x%x Cfing=0x%x Axes=0x%x X=%d Y=%d Z=%d Rz=%d Area=%u Hx=%.1f Hy=%.1f \n",
			bimanualname, handname,handeventname,
           pHandMotion->selectedChord.word, pHandMotion->currentFingers.fields.fingers,
		   pHandMotion->slideAxesCode,
           pHandMotion->Xtrans,pHandMotion->Ytrans,pHandMotion->Ztrans,pHandMotion->Zrot,
		   pHandMotion->handPressure,
		   pHandMotion->handPositionX,pHandMotion->handPositionY);
}

