/**
 *  @file FWHID_HandMotion.h
 *  FingerWorks HID Hand Motion SDK (External API)
 *
 *  @author Created by Wayne Westerman on Wed Jun 25 2003.
  *  @author Last Updated by James Orr on Tuesday Sep 7 2004.

  *  v0.65 fixed FWHID_ZoneSpecifier for little endian  
  *  v1.0 changed FWHID_ZoneSpecifier to FWHID_FingerSpecifier, and added handActionCode and handPositionX/Y
  *  Copyright (c) 2003-2004 FingerWorks, Inc. All rights reserved. 
  **/
 /* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is FWHID_HandMotion.h, 
 *   External API for FingerWorks HID Hand Motion SDK.
 *
 * The Initial Developer of the Original Code is
 * Wayne Westerman, FingerWorks, Inc.
 *
 * Portions created by the Initial Developer are Copyright (C) 2003-2004
 * by FingerWorks Inc. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */
 

#ifndef _FWHID_HandMotion_h_
#define _FWHID_HandMotion_h_

#include "FWHID_PlatformDefs.h"

#ifdef __cplusplus
extern "C" {
#endif
    
/** "Any fingertip permutation" chords.
    These are the chords currently defined in the MyGesture Editor libraries.

    Thumb presence, fingertip count, and neutral/spread finger arrangement 
    are recognized reliably from informed operators
    but actual permutation of palms and starting fingertips may vary.
    
    See FWHID_ZoneSpecifier for ongoing finger identity estimates
Key:
   T = thumb
   #F = fingertip count
   S = spread (not neutral) hand arrangement
*/
#define kFWHID_Chord_1F 0x1004
#define kFWHID_Chord_2F 0x100C
#define kFWHID_Chord_2FS 0x1812
#define kFWHID_Chord_3F 0x100E
#define kFWHID_Chord_3FS 0x181C
#define kFWHID_Chord_4F 0x101E
#define kFWHID_Chord_4FS 0x181E
#define kFWHID_Chord_T1F 0x1005
#define kFWHID_Chord_T1FS 0x1805
#define kFWHID_Chord_T2F 0x1007
#define kFWHID_Chord_T2FS 0x1819
#define kFWHID_Chord_TIPS 0x1813
#define kFWHID_Chord_T3F 0x100F
#define kFWHID_Chord_T3FS 0x181D
#define kFWHID_Chord_T4F 0x101F
#define kFWHID_Chord_T4FS 0x181F
#define kFWHID_Chord_PenGrip 0x1100
#define kFWHID_Chord_TriPenGrip 0x1200
#define kFWHID_Chord_FistPalms 0x1400

#define kFWHID_RightHandMask 0x8000
#define kFWHID_LeftHandMask 0x4000
#define kFWHID_BimanualSyncMask 0x2000
#define kFWHID_AnyFingertipPermutationMask 0x1000

#define kFWHID_ActionCode_WinHide 0x81
#define kFWHID_ActionCode_WinSendToBottom 0x82
#define kFWHID_ActionCode_WinBringToTop 0x83
#define kFWHID_ActionCode_WinMinimize 0x84
#define kFWHID_ActionCode_WinRestore 0x85
#define kFWHID_ActionCode_WinToggle 0x86
#define kFWHID_ActionCode_WinMaximize 0x87

#define kFWHID_ActionCode_SubWinSendToTop 0x92
#define kFWHID_ActionCode_SubWinSendToBottom 0x93
#define kFWHID_ActionCode_SubWinToggle 0x94
#define kFWHID_ActionCode_SubWinMaximize 0x95
#define kFWHID_ActionCode_SubWinMinimize 0x96
#define kFWHID_ActionCode_SubWinRestore 0x97
//create new Carbon event class 'hand'?
enum {
    kEventClassHand = 'hand'
};

//'hand' event class sub-types
enum {
    kEventHandNoChange = 0,
    kEventHandChordSelected = 1, //should not include any motion
    kEventHandChordReleased = 2, //should not include any motion
	kEventHandCommand = 4,
    kEventHandPressure = 8,
    kEventHandTranslate = 9,
    kEventHandRotate = 10,
    kEventHandTranslateRotate = 11,
    kEventHandScale = 12,
    kEventHandTranslateScale = 13,
    kEventHandRotateScale = 14,
    kEventHandMotion = 15, //any/all of translation/rotation/scaling!
    kEventHandIDmax
};

/* FingerWorks products allow chord selection, i.e. update the FWHID_ChordSpecifier,
   only at initial hand touchdown. Further changes in finger presence after initial
   touchdown are ignored for chord selection purposes until the whole hand lifts. */   
#ifdef __BIG_ENDIAN__
typedef union {
    unsigned short word;
    struct {
        //PowerPC & Sun compilers put bit 15 at top of bitfield, bit 0 at bottom!
        unsigned int lefthand : 1;  //bit 15
        unsigned int righthand : 1;
        unsigned int bimanualSync : 1;
        unsigned int anyFingertipPermutation : 1;
        unsigned int spread : 1;
        unsigned int fist : 1;
        unsigned int triPenGrip : 1;
        unsigned int penGrip : 1;
        unsigned int forePalms : 1;
        unsigned int innerPalm : 1;
        unsigned int outerPalm : 1;
        unsigned int pinky : 1;
        unsigned int ring : 1;
        unsigned int middle : 1;
        unsigned int index : 1;
        unsigned int thumb : 1; //bit0
    } flags;
    struct {
        //PowerPC compilers put bit 15 at top of bitfield, bit 0 at bottom!
        unsigned int hand : 3;
        unsigned int anyFingertipPermutation : 1;
        unsigned int grip : 4;
        unsigned int fingers : 8;   
    } fields;
    struct {
        //PowerPC compilers put bit 15 at top of bitfield, bit 0 at bottom!
        unsigned int hand : 3;
        unsigned int chord : 13;   
    } compare;
} FWHID_ChordSpecifier;

#else //Intel architectures are little Endian

typedef union {
    unsigned short word;
    struct {
        //Intel compilers put bit 0 at top of bitfield, bit 15 at bottom!
        unsigned int thumb : 1; //bit0
        unsigned int index : 1;
        unsigned int middle : 1;
        unsigned int ring : 1;
        unsigned int pinky : 1;
        unsigned int outerPalm : 1;
        unsigned int innerPalm : 1;
        unsigned int forePalms : 1;
        unsigned int penGrip : 1;
        unsigned int triPenGrip : 1;
        unsigned int fist : 1;
        unsigned int spread : 1;
        unsigned int anyFingertipPermutation : 1;
        unsigned int bimanualSync : 1;
        unsigned int righthand : 1;
        unsigned int lefthand : 1;  //bit 15
    } flags;
    struct {
        //Intel compilers put bit 0 at top of bitfield, bit 15 at bottom!
        unsigned int fingers : 8;   
        unsigned int grip : 4;
        unsigned int anyFingertipPermutation : 1;
        unsigned int hand : 3;
    } fields;
    struct {
        //Intel compilers put bit 0 at top of bitfield, bit 15 at bottom!
        unsigned int chord : 13;   
        unsigned int hand : 3;
    } compare;
} FWHID_ChordSpecifier;
#endif //POWERPC

/* FingerWorks products continually update the FWHID_ZoneSpecifier.
The zone fields roughly indicate hand position on the surface with
respect to 2cm zones around home row.  A hand within +-1cm of home
row (or surface center) is in zone 0.  Above home row (+1 to +3cm)
is vertical zone 1, below home row (-1 to -3cm) is vertical zone -1,
and so on for zones -7 thru +7, horizontal and vertical.  

FWHID_ZoneSpecifier also includes continuosly updated estimates
of which fingers are actively touching the surface. These are
estimates, not guaranteed to be accurate, but thumb presence,
palm presence, and fingertip count are pretty reliable.  */   

#ifdef __BIG_ENDIAN__
typedef union {
    unsigned short word;
    struct {
        //PowerPC compilers put bit 15 at top of bitfield, bit 0 at bottom!
		unsigned int reserved : 8;
        unsigned int forePalms : 1;
        unsigned int innerPalm : 1;
        unsigned int outerPalm : 1;
        unsigned int pinky : 1;
        unsigned int ring : 1;
        unsigned int middle : 1;
        unsigned int index : 1;
        unsigned int thumb : 1; //bit0
    } flags;
    struct {
        //PowerPC compilers put bit 15 at top of bitfield, bit 0 at bottom!
		unsigned int reserved : 8;
        unsigned int fingers : 8;   
    } fields;
} FWHID_FingerSpecifier;

#else //Intel architectures are little Endian

typedef union {
    unsigned short word;
    struct {
        //Intel compilers put bit 0 at top of bitfield, bit 15 at bottom!
        unsigned int thumb : 1; //bit0
        unsigned int index : 1;
        unsigned int middle : 1;
        unsigned int ring : 1;
        unsigned int pinky : 1;
        unsigned int outerPalm : 1;
        unsigned int innerPalm : 1;
        unsigned int forePalms : 1;
		unsigned int reserved : 8;
    } flags;
    struct {
        //Intel compilers put bit 0 at top of bitfield, bit 15 at bottom!
        unsigned int fingers : 8;   
		unsigned int reserved : 8;
    } fields;
} FWHID_FingerSpecifier;

#endif //endianess

typedef struct {
    FWHID_ChordSpecifier selectedChord;
    FWHID_ChordSpecifier lastChord;
    FWHID_FingerSpecifier currentFingers; 
    FWHID_FingerSpecifier lastFingers;
    unsigned handActionCode;
    unsigned handPressure;
    unsigned lastPressure;
	/**In centimeters at 25 um resolution
	@brief The X coordinate of the hand */
	float handPositionX; 
	/**In centimeters at 25 um resolution
	@brief The Y coordinate of the hand */
	float handPositionY; //centimeters at 25 um resolution
    int Xtrans;
    int Ytrans;
    int Ztrans;//'zoom'
    int Xrot; /**<@brief this axis inactive as of FingerWorks Firmware v1.33*/
    int Yrot; /**<@brief this axis inactive as of FingerWorks Firmware v1.33*/
    int Zrot; /**<@brief rotate*/
    int Xscale;         /**<@brief this axis inactive as of FingerWorks Firmware v1.33*/
    int Yscale;         /**<@brief this axis inactive as of FingerWorks Firmware v1.33*/
    int Zscale;         /**<@brief this axis inactive as of FingerWorks Firmware v1.33*/
	unsigned slideAxesCode;
} FWHID_HandMotion, *FWHID_HandMotionPtr;


/*  NOTE:  Receiver does not 'own' the data pointed to by handMotionDatap!
   It's a stack-allocated copy.  */
typedef void (*FWHIDHandEventCALLBACK)
(int handEventID, FWHID_HandMotionPtr pHandMotion);


//NOTE:  Your Windows App's event processing loop must also call FWHID_WndProc() (see FWHID_PlatformDefs.h) as a default processor for WM_INPUT and/or WM_DEVICE_CHANGED messages

//your application should must set a callback so it can receive hand events!
void FWHID_setHandEventCallback(FWHIDHandEventCALLBACK fwcallback);
FWHIDHandEventCALLBACK FWHID_getHandEventCallback(void);

//this default hand event callback just prints the hand motion event to stdout (console)
void FWHID_printHandMotion(int handEventID, FWHID_HandMotionPtr pHandMotion);

int FWHID_snprintfHandMotion(char *buffer, size_t bufsize, int handEventID, FWHID_HandMotionPtr pHandMotion);

#ifdef __cplusplus
}
#endif

#endif // _FWHID_HandMotion_h_

