/*
 *  XWinderCalcs.c
 *  FingerWorks XWinder window manipulation utility 
 *    Platform-independent hand/window motion calculations
 *
 *  Created by Wayne Westerman on Sunday Nov 9 2003.
 *  Last Updated by Wayne Westerman on Wednesday Feb 11 2004.
 *  Copyright (c) 2003-2004 FingerWorks, Inc. All rights reserved. 
 **/
/* ***** BEGIN LICENSE BLOCK *****
* Version: MPL 1.1/GPL 2.0/LGPL 2.1
*
* The contents of this file are subject to the Mozilla Public License Version
* 1.1 (the "License"); you may not use this file except in compliance with
* the License. You may obtain a copy of the License at
* http://www.mozilla.org/MPL/
*
* Software distributed under the License is distributed on an "AS IS" basis,
* WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
* for the specific language governing rights and limitations under the
* License.
*
* The Original Code is XWinderCalcs.c, 
*   Platform-independent hand/window motion calculations for XWinder window manipulation utility.
*
* The Initial Developer of the Original Code is
* Wayne Westerman, FingerWorks, Inc.
*
*	v1.05   Wrapped XWFWhandleActionCode() with XWFWgrabWindow() and XWFWdropGrabbedWindow() so action code events can work without hand select/release wrappers
*
* Portions created by the Initial Developer are Copyright (C) 2003
* by FingerWorks Inc. All Rights Reserved.
*
* Contributor(s):
*
* Alternatively, the contents of this file may be used under the terms of
* either the GNU General Public License Version 2 or later (the "GPL"), or
* the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
* in which case the provisions of the GPL or the LGPL are applicable instead
* of those above. If you wish to allow use of your version of this file only
* under the terms of either the GPL or the LGPL, and not to allow others to
* use your version of this file under the terms of the MPL, indicate your
* decision by deleting the provisions above and replace them with the notice
* and other provisions required by the GPL or the LGPL. If you do not delete
* the provisions above, a recipient may use your version of this file under
* the terms of any one of the MPL, the GPL or the LGPL.
*
* ***** END LICENSE BLOCK ***** */


#include "XWinderCalcs.h"

#ifdef __cplusplus
extern "C" {
#endif

FWHID_HandMotion lefthandPending;
FWHID_HandMotion righthandPending;

// any shorter than 80ms and ~1GHz systems can't keep up on redraw of complex MDI's like Visual Studio IDE
static XWFWAbsoluteTime tlast_wind = 0;

/** Converts accumulated hand motion to a change in window bounds.  The windowing system may
    reject the bounds change if that would put the window out of range, but this function doesn't
    really care. */
static XWFWRect * xwfw_handMotion2winMotion(XWFWRect *bounds_changep, FWHID_HandMotionPtr onehandp) {
    static XWFWAbsoluteTime tlast_move_lock = 0;
    //assume zero change default
    bounds_changep->origin.x = 0;
    bounds_changep->origin.y = 0;
    bounds_changep->size.width = 0;
    bounds_changep->size.height = 0;
    if(lefthandPending.selectedChord.flags.bimanualSync && righthandPending.selectedChord.flags.bimanualSync) {
        if(XWFWgetBooleanPreference("HoldBothCorners") == true) {
            //move upper left corner by lefthand motion
            // for window to move without resizing righthand should move same amount!
            bounds_changep->origin.x = (float)lefthandPending.Xtrans;
            bounds_changep->origin.y = (float)lefthandPending.Ytrans;
            //resize window by difference in hand translations
            // this makes lower right corner follow right hand
            bounds_changep->size.width = (float)(righthandPending.Xtrans-lefthandPending.Xtrans);
            bounds_changep->size.height = (float)(righthandPending.Ytrans-lefthandPending.Ytrans);
            
            if((lefthandPending.Xtrans*righthandPending.Xtrans > 0 && lefthandPending.Ytrans*righthandPending.Ytrans > 0)
               || (lefthandPending.Xtrans*righthandPending.Xtrans > 0
                   && fabs(bounds_changep->origin.x) > 1.5*fabs(bounds_changep->size.height))
               || (lefthandPending.Ytrans*righthandPending.Ytrans > 0
                   && fabs(bounds_changep->origin.y) > 1.5*fabs(bounds_changep->size.width)) 
               ) {
                //suppress resize when both hands moving in the same direction
                tlast_move_lock = XWFWgetAbsoluteTime();
                bounds_changep->size.width = 0;
                bounds_changep->size.height = 0;
            } else if((fabs(bounds_changep->size.width) < 15 
					   && fabs(bounds_changep->size.height) < 15
                       && (XWFWgetAbsoluteTime() - tlast_move_lock) < 2)) {
                //suppress small size changes when we've only been moving recently
                bounds_changep->size.width = 0;
                bounds_changep->size.height = 0;
            } else if(lefthandPending.Xtrans*righthandPending.Xtrans != 0
					|| lefthandPending.Ytrans*righthandPending.Ytrans != 0){
				//both hands must be moving in opposite directions
                //allow both position and size changes to pass through and clear tlast_move_lock
                tlast_move_lock = 0;
            }    
        } else {
            //lefthand changes window origin (moves window)
            bounds_changep->origin.x = (float)lefthandPending.Xtrans;
            bounds_changep->origin.y = (float)lefthandPending.Ytrans;
            //righthand changes window size (resizes window)
            bounds_changep->size.width = (float)righthandPending.Xtrans;
            bounds_changep->size.height = (float)righthandPending.Ytrans;
        }
    } else if(onehandp != NULL) {
		bool default_three_fingers = true;
		bool orthoStretchAsDefault = XWFWgetBooleanPreference("OrthoStretchAsDefault");
		if(onehandp->currentFingers.flags.index
           && onehandp->currentFingers.flags.middle
           && onehandp->currentFingers.flags.ring
		   && onehandp->currentFingers.flags.pinky) {
			   //all four fingertips now touching
			default_three_fingers = false;
		}
        //hand translation moves window
        bounds_changep->origin.x = (float)onehandp->Xtrans;
        bounds_changep->origin.y = (float)onehandp->Ytrans;

        if((default_three_fingers && !orthoStretchAsDefault)
		   || (!default_three_fingers && orthoStretchAsDefault)) {
            //hand scaling varies width and height equally, 
            //   while rotation varies aspect ratio (of width to height) with double rotational sensitivity to avoid drastic hand twists
            bounds_changep->size.width = (float)(onehandp->Ztrans - 2*onehandp->Zrot);
            bounds_changep->size.height = (float)(onehandp->Ztrans + 2*onehandp->Zrot);
        } else { 
            //hand scaling varies window width
            bounds_changep->size.width = (float)onehandp->Ztrans;
            //hand rotation varies window height with double rotational senitivity to avoid drastic hand twists
            bounds_changep->size.height = (float)2*onehandp->Zrot;
        }
    }
    return bounds_changep;
}
                     
/** Clears accumulated motion from either or both lefthandPending, righthandPending.
    Called after xwfw_handMotion2winMotion() conversion is done.  */
static void xwfw_clearPendingMotion(FWHID_HandMotionPtr pendingSourceHandp) {
    if(lefthandPending.selectedChord.flags.bimanualSync
       && righthandPending.selectedChord.flags.bimanualSync) {
        //clear both right and left-hand pending motions
        lefthandPending.Xtrans=0;
        lefthandPending.Ytrans=0;
        lefthandPending.Ztrans=0;
        lefthandPending.Zrot=0;
        righthandPending.Xtrans=0;
        righthandPending.Ytrans=0;
        righthandPending.Ztrans=0;
        righthandPending.Zrot=0;
    } else if(pendingSourceHandp != NULL) {
        //clear source hand only
        pendingSourceHandp->Xtrans=0;
        pendingSourceHandp->Ytrans=0;
        pendingSourceHandp->Ztrans=0;
        pendingSourceHandp->Zrot=0;
    }
}


/** Computes window bounds change, attempts to change the bounds, and clears accumulated motion. */
static void xwfw_windGrabbedWindow(FWHID_HandMotionPtr pendingSourceHandp) {
    XWFWRect by_amount = {{0,0},{0,0}};

    xwfw_handMotion2winMotion(&by_amount, pendingSourceHandp);
    
    XWFWchangeGrabbedWindowBounds(&by_amount, pendingSourceHandp);
    
    xwfw_clearPendingMotion(pendingSourceHandp);

    tlast_wind = XWFWgetAbsoluteTime();
}



/** The central hand motion event handler used as a callback from FWHID_HandMotion library.
    Dispatches XWFWgrabWindow(), XWFWdropGrabbedWindow(), and xwfw_windGrabbedWindow()
    depending on event type.
*/

void XWFWhandleHandMotion(int handEventID, FWHID_HandMotionPtr pHandMotion) {
    static int gXWFWhandmotion_calls = 0;
	//	FWHID_printHandMotion(handEventID,pHandMotion);        
    if(pHandMotion != NULL) {
        FWHID_HandMotionPtr pendingSourceHandp = NULL;
        if(pHandMotion->selectedChord.flags.righthand) {
            pendingSourceHandp = &righthandPending;
        } else if(pHandMotion->selectedChord.flags.lefthand) {
            pendingSourceHandp = &lefthandPending;
        }
        //copy 
        if(pHandMotion->selectedChord.flags.righthand) {
            righthandPending.selectedChord.word = pHandMotion->selectedChord.word;
            righthandPending.currentFingers.word = pHandMotion->currentFingers.word;
        }
        if(pHandMotion->selectedChord.flags.lefthand) {
            lefthandPending.selectedChord.word = pHandMotion->selectedChord.word;
            lefthandPending.currentFingers.word = pHandMotion->currentFingers.word;
        }
        if(handEventID == kEventHandChordSelected) {
            if(pHandMotion->selectedChord.compare.chord == kFWHID_Chord_TIPS
               || pHandMotion->selectedChord.compare.chord == kFWHID_Chord_T2F
               || ((pHandMotion->selectedChord.compare.chord == kFWHID_Chord_3F
                    || pHandMotion->selectedChord.compare.chord == kFWHID_Chord_T3F)
                   //bimanual only!
                   && pHandMotion->selectedChord.flags.bimanualSync
                   //only respond to the first of two hand's chord selections to arrive
                   //once second arrives they'll be the same!
                   && lefthandPending.selectedChord.compare.chord != righthandPending.selectedChord.compare.chord)
               ) {
                //grab (highlight) the window under pointer
                XWFWgrabWindow(pendingSourceHandp);
            }
        } else if(handEventID == kEventHandChordReleased) {
            if(!pHandMotion->selectedChord.flags.bimanualSync
               //for bimanual slides must wait for both hands to release
               || (lefthandPending.selectedChord.compare.chord == 0 && righthandPending.selectedChord.compare.chord == 0)) {
                //release (unhighlight) the window for this hand
                XWFWdropGrabbedWindow(pendingSourceHandp);
                //FWHID_printHandMotion(handEventID,pHandMotion);        
            } 
        } else if(pHandMotion->selectedChord.compare.chord == kFWHID_Chord_4F
					&& !pHandMotion->selectedChord.flags.bimanualSync){
            XWFWpanWindow(pHandMotion->Ztrans);			
		} else if(handEventID == kEventHandCommand) {
			//do we want to force a new window grab whenever we get a handActionCode, or just rely on ChordSelect/Release
//                XWFWgrabWindow(pendingSourceHandp); 
			if(pHandMotion->handActionCode >= kFWHID_ActionCode_WinMinimize
				&& pHandMotion->handActionCode <= kFWHID_ActionCode_WinMaximize) {
				XWFWgrabWindow(pendingSourceHandp);
				XWFWhandleActionCode(pHandMotion->handActionCode);
				XWFWdropGrabbedWindow(pendingSourceHandp);
			}
        } else if(handEventID > kEventHandPressure) {
			XWFWTimeInterval refresh_interval_s = 0;
			if(pHandMotion->selectedChord.flags.bimanualSync)
				refresh_interval_s = XWFWgetDoublePreference("BimanualRefreshDelay");
			else refresh_interval_s = XWFWgetDoublePreference("UnimanualRefreshDelay");
            gXWFWhandmotion_calls++;
            if(pendingSourceHandp) {
                //accumulate motion from incoming event into pendingSourceHandp
                pendingSourceHandp->Xtrans += pHandMotion->Xtrans;
                pendingSourceHandp->Ytrans += pHandMotion->Ytrans;
                pendingSourceHandp->Ztrans += pHandMotion->Ztrans;
                pendingSourceHandp->Zrot += pHandMotion->Zrot;
            }
               
            if(XWFWgetAbsoluteTime() - tlast_wind > refresh_interval_s
				/*&& (!pHandMotion->selectedChord.flags.bimanualSync
					|| (gXWFWhandmotion_calls % 2) == 0)*/) {
                //apply hand motion right away
                XWFWstopPostponedWindTimer();
                xwfw_windGrabbedWindow(pendingSourceHandp);
            } else {
                //wait a bit to see if other hand will supply motion in bimanual manipulation
                //  if other hand doens, comparing hand motions will help us suppress minor resize fluctuations
                XWFWstartPostponedWindTimer(xwfw_windGrabbedWindow,pendingSourceHandp,refresh_interval_s);
            }
        }                                
    }
    //FWHID_printHandMotion(handEventID,pHandMotion);        
}

#ifdef __cplusplus
}
#endif
